{=========================================================================
  RkAddr Unit                                                               
                                                                            
  This unit implements the RkAddress component which is comprised of the
  the following edit fields: First Name, Last Name, Street, City, and ZIP.
  The State field is actually a combo box which is populated with the 50
  states including the District of Columbia. The edit fields are data-
  aware, and thus this component can be hooked up to a DataSource.

  Developing Custom Delphi 3 Components
  Copyright  1995-1997 by Ray Konopka
=========================================================================}

{$I DCDC.INC}

unit RkAddr;

interface

uses
  Classes, Messages, Controls, StdCtrls, DB, DBCtrls, Graphics,
  ExtCtrls, RkCommon;

type
  TLabelPosition = ( lpOnLeft, lpOnTop );
  TEditField = ( efFirstName, efLastName, efStreet, efCity, efZIP );
  TEditChangeEvent = procedure(Field: TEditField; Text: string) of object;

  TRkAddress = class( TWinControl )
  private
    FAboutInfo : TRkAboutInfo;
    FEdtFirstName : TDBEdit;
    FLblFirstName : TLabel;
    FEdtLastName : TDBEdit;
    FLblLastName : TLabel;
    FEdtStreet : TDBEdit;
    FLblStreet : TLabel;
    FEdtCity : TDBEdit;
    FLblCity : TLabel;
    FCbxState : TDBComboBox;
    FLblState : TLabel;
    FEdtZIP : TDBEdit;
    FLblZIP : TLabel;
    FStateList : TStrings;    // Defaults to a list of state abbreviations
    FLabelPosition : TLabelPosition;
    FOnChange : TEditChangeEvent;      // Change event for all edit fields

    function GetCharCase : TEditCharCase;
    procedure SetCharCase( Value : TEditCharCase );
    function GetField( Index : Integer ) : string;
    procedure SetField( Index : Integer; Value : string );
    function GetCaption( Index : Integer ) : TCaption;
    procedure SetCaption( Index : Integer; Value : TCaption );
    function GetFieldColor : TColor;
    procedure SetFieldColor( Value : TColor );

    function GetDataSource : TDataSource;
    procedure SetDataSource( Value : TDataSource );
    procedure SetLabelPosition( Value : TLabelPosition );
    procedure SetStateList( Value : TStrings );

    function CreateEdit : TDBEdit;
    function CreateLabel( ACaption : TCaption ) : TLabel;
    function CreateCombo : TDBComboBox;
    procedure CreateStateList;
    procedure DoChange( Sender : TObject );
    function GetFontHeight( Font : TFont ) : Integer;
    procedure WMSize( var Msg : TWMSize ); message wm_Size;
    procedure CMFontChanged( var Msg : TMessage ); message cm_FontChanged;
  protected
    procedure Change( Field : TEditField; Text : string ); dynamic;
    procedure CreateWnd; override;
    procedure PositionControls; virtual;
    procedure Resize; virtual;
  public
    constructor Create( AOwner : TComponent ); override;
    destructor Destroy; override;

    property EdtFirstName : TDBEdit
      read FEdtFirstName;

  published
    property About : TRkAboutInfo
      read FAboutInfo
      write FAboutInfo
      stored False;

    property CharCase : TEditCharCase
      read GetCharCase
      write SetCharCase;

    property FieldColor : TColor
      read GetFieldColor
      write SetFieldColor;

    property DataSource : TDataSource
      read GetDataSource
      write SetDataSource;

    property FirstNameCaption : TCaption
      index 1
      read GetCaption
      write SetCaption;

    property LastNameCaption : TCaption
      index 2
      read GetCaption
      write SetCaption;

    property StreetCaption : TCaption
      index 3
      read GetCaption
      write SetCaption;

    property CityCaption : TCaption
      index 4
      read GetCaption
      write SetCaption;

    property StateCaption : TCaption
      index 5
      read GetCaption
      write SetCaption;

    property ZIPCaption : TCaption
      index 6
      read GetCaption
      write SetCaption;

    property FirstNameField : string
      index 1
      read GetField
      write SetField;

    property LastNameField : string
      index 2
      read GetField
      write SetField;

    property StreetField : string
      index 3
      read GetField
      write SetField;

    property CityField : string
      index 4
      read GetField
      write SetField;

    property StateField : string
      index 5
      read GetField
      write SetField;

    property ZIPField : string
      index 6
      read GetField
      write SetField;

    property LabelPosition : TLabelPosition
      read FLabelPosition
      write SetLabelPosition
      default lpOnTop;

    property StateList : TStrings
      read FStateList
      write SetStateList;

    property OnChange : TEditChangeEvent
      read FOnChange
      write FOnChange;

    property Font;
    property ParentFont;
  end;


implementation

uses
  Windows;

{$IFNDEF VER90}
{$IFNDEF CHAPTER_10}
resourcestring
  SAddrFirstName = 'First Name';
  SAddrLastName  = 'Last Name';
  SAddrStreet    = 'Street';
  SAddrCity      = 'City';
  SAddrState     = 'State';
  SAddrZIP       = 'ZIP';
{$ENDIF}
{$ENDIF}

{========================}
{== TRkAddress Methods ==}
{========================}


constructor TRkAddress.Create( AOwner : TComponent );
begin
  inherited Create( AOwner );

  // Don't allow other controls to be dropped onto the address component
  ControlStyle := ControlStyle - [csAcceptsControls ];

  // Clean up is handled when the TComponent ancestor class frees
  // all components on the Components list.



{$IFDEF VER90}

  FLblFirstName := CreateLabel( 'First Name' );
  FEdtFirstName := CreateEdit;

  FLblLastName := CreateLabel( 'Last Name' );
  FEdtLastName := CreateEdit;

  FLblStreet := CreateLabel( 'Street' );
  FEdtStreet := CreateEdit;

  FLblCity := CreateLabel( 'City' );
  FEdtCity := CreateEdit;

  FLblState := CreateLabel( 'State' );
  FCbxState := CreateCombo;

  FLblZIP := CreateLabel( 'ZIP' );
  FEdtZIP := CreateEdit;

{$ELSE}

  {$IFNDEF CHAPTER_10}

  FLblFirstName := CreateLabel( SAddrFirstName );
  FEdtFirstName := CreateEdit;

  FLblLastName := CreateLabel( SAddrLastName );
  FEdtLastName := CreateEdit;

  FLblStreet := CreateLabel( SAddrStreet );
  FEdtStreet := CreateEdit;

  FLblCity := CreateLabel( SAddrCity );
  FEdtCity := CreateEdit;

  FLblState := CreateLabel( SAddrState );
  FCbxState := CreateCombo;

  FLblZIP := CreateLabel( SAddrZIP );
  FEdtZIP := CreateEdit;

  {$ELSE}

  FLblFirstName := CreateLabel( 'First Name' );
  FEdtFirstName := CreateEdit;

  FLblLastName := CreateLabel( 'Last Name' );
  FEdtLastName := CreateEdit;

  FLblStreet := CreateLabel( 'Street' );
  FEdtStreet := CreateEdit;

  FLblCity := CreateLabel( 'City' );
  FEdtCity := CreateEdit;

  FLblState := CreateLabel( 'State' );
  FCbxState := CreateCombo;

  FLblZIP := CreateLabel( 'ZIP' );
  FEdtZIP := CreateEdit;

  {$ENDIF}

{$ENDIF}

  CreateStateList;

  Width := 382;
  Height := 78;

  LabelPosition := lpOnTop;
  PositionControls;
end; {= TRkAddress.Create =}


destructor TRkAddress.Destroy;
begin
  // No need to destroy the embedded components, because the inherited
  // destructor will destroy all components on the Components list.
  // FStateList is not on that list, so we need to destroy it here.
  FStateList.Free;
  inherited Destroy;
end;


function TRkAddress.CreateLabel( ACaption : TCaption ) : TLabel;
begin
  Result := TLabel.Create( Self );
  Result.Parent := Self;
  Result.Visible := True;
  Result.Caption := ACaption;
end;


function TRkAddress.CreateEdit : TDBEdit;
begin
  Result := TDBEdit.Create( Self );
  Result.Parent := Self;
  Result.Visible := True;

  // Assign OnChange event of each Edit field to point to
  // TRkAddress.DoChange method
  Result.OnChange := DoChange;
end;


function TRkAddress.CreateCombo : TDBComboBox;
begin
  Result := TDBComboBox.Create( Self );
  Result.Parent := Self;
  Result.Visible := True;
  Result.Sorted := True;
end;


procedure TRkAddress.CreateWnd;
begin
  inherited CreateWnd;

  // When CreateWnd is called, the Items list of FCbxState is cleared.
  // Therefore, the contents of the FStateList are copied back into
  // FCbxState.

  // Plus, you cannot add items to a ComboBox until its Parent's window
  // handle has been allocated. Therefore, this code also serves to
  // initially populate the ComboBox with state names.
  FCbxState.Items.Assign( FStateList );
end;


procedure TRkAddress.CreateStateList;
begin
  FStateList := TStringList.Create;
  FStateList.BeginUpdate;
  try
    FStateList.Add( 'AK' );      // Alaska
    FStateList.Add( 'AL' );      // Alabama
    FStateList.Add( 'AR' );      // Arkansas
    FStateList.Add( 'AZ' );      // Arizona
    FStateList.Add( 'CA' );      // California
    FStateList.Add( 'CO' );      // Colorado
    FStateList.Add( 'CT' );      // Connecticut
    FStateList.Add( 'DC' );      // District of Columbia
    FStateList.Add( 'DE' );      // Delaware
    FStateList.Add( 'FL' );      // Florida
    FStateList.Add( 'GA' );      // Georgia
    FStateList.Add( 'HI' );      // Hawaii
    FStateList.Add( 'IA' );      // Iowa
    FStateList.Add( 'ID' );      // Idaho
    FStateList.Add( 'IL' );      // Illinois
    FStateList.Add( 'IN' );      // Indiana
    FStateList.Add( 'KS' );      // Kansas
    FStateList.Add( 'KY' );      // Kentucky
    FStateList.Add( 'LA' );      // Louisiana
    FStateList.Add( 'MA' );      // Massachusetts
    FStateList.Add( 'MD' );      // Maryland
    FStateList.Add( 'ME' );      // Maine
    FStateList.Add( 'MI' );      // Michigan
    FStateList.Add( 'MN' );      // Minnesota
    FStateList.Add( 'MO' );      // Missouri
    FStateList.Add( 'MS' );      // Mississippi
    FStateList.Add( 'MT' );      // Montana
    FStateList.Add( 'NC' );      // North Carolina
    FStateList.Add( 'ND' );      // North Dakota
    FStateList.Add( 'NE' );      // Nebraska
    FStateList.Add( 'NH' );      // New Hampshire
    FStateList.Add( 'NJ' );      // New Jersey
    FStateList.Add( 'NM' );      // New Mexico
    FStateList.Add( 'NV' );      // Nevada
    FStateList.Add( 'NY' );      // New York
    FStateList.Add( 'OH' );      // Ohio
    FStateList.Add( 'OK' );      // Oklahoma
    FStateList.Add( 'OR' );      // Oregon
    FStateList.Add( 'PA' );      // Pennsylvania
    FStateList.Add( 'RI' );      // Rhode Island
    FStateList.Add( 'SC' );      // South Carolina
    FStateList.Add( 'SD' );      // South Dakota
    FStateList.Add( 'TN' );      // Tennessee
    FStateList.Add( 'TX' );      // Texas
    FStateList.Add( 'UT' );      // Utah
    FStateList.Add( 'VA' );      // Virginia
    FStateList.Add( 'VT' );      // Vermont
    FStateList.Add( 'WA' );      // Washington
    FStateList.Add( 'WI' );      // Wisconsin
    FStateList.Add( 'WV' );      // West Virginia
    FStateList.Add( 'WY' );      // Wyoming
  finally
    FStateList.EndUpdate;
  end;
end; {= TRkAddress.CreateStateList =}


procedure TRkAddress.SetStateList( Value : TStrings );
begin
  FStateList.Assign( Value );
end;


procedure TRkAddress.Change( Field : TEditField; Text : string );
begin
  if Assigned( FOnChange ) then
    FOnChange( Field, Text );
end;


{=========================================================================
  TRkAddress.DoChange

  This method gets called if the OnChange event occurs for any of the
  edit fields contained in this component.  The Change event dispatch
  method is called to surface those events to the user.
=========================================================================}

procedure TRkAddress.DoChange( Sender : TObject );
var
  Field : TEditField;
begin
  if Sender = FEdtFirstName then
    Field := efFirstName
  else if Sender = FEdtLastName then
    Field := efLastName
  else if Sender = FEdtStreet then
    Field := efStreet
  else if Sender = FEdtCity then
    Field := efCity
  else
    Field := efZIP;
  Change( Field, TDBEdit( Sender ).Text );
end;


function TRkAddress.GetCharCase : TEditCharCase;
begin
  Result := FEdtFirstName.CharCase;
end;


procedure TRkAddress.SetCharCase( Value : TEditCharCase );
begin
  if Value <> FEdtFirstName.CharCase then
  begin
    FEdtFirstName.CharCase := Value;
    FEdtLastName.CharCase := Value;
    FEdtStreet.CharCase := Value;
    FEdtCity.CharCase := Value;
    FEdtZIP.CharCase := Value;
  end;
end;


function TRkAddress.GetFieldColor : TColor;
begin
  Result := FEdtFirstName.Color;
end;


procedure TRkAddress.SetFieldColor( Value : TColor );
begin
  if Value <> FEdtFirstName.Color then
  begin
    FEdtFirstName.Color := Value;
    FEdtLastName.Color := Value;
    FEdtStreet.Color := Value;
    FEdtCity.Color := Value;
    FCbxState.Color := Value;
    FEdtZIP.Color := Value;
  end;
end;


function TRkAddress.GetDataSource : TDataSource;
begin
  // Use FEdtFirstName to get current DataSource
  Result := FEdtFirstName.DataSource;
end;


procedure TRkAddress.SetDataSource( Value : TDataSource );
begin
  if Value <> FEdtFirstName.DataSource then
  begin
    // Assign All Internal Controls to Same DataSource
    FEdtFirstName.DataSource := Value;
    FEdtLastName.DataSource := Value;
    FEdtStreet.DataSource := Value;
    FEdtCity.DataSource := Value;
    FCbxState.DataSource := Value;
    FEdtZIP.DataSource := Value;
  end;
end;


function TRkAddress.GetField( Index : Integer ) : string;
begin
  case Index of
    1: Result := FEdtFirstName.DataField;
    2: Result := FEdtLastName.DataField;
    3: Result := FEdtStreet.DataField;
    4: Result := FEdtCity.DataField;
    5: Result := FCbxState.DataField;
    6: Result := FEdtZIP.DataField;
  end;
end;


procedure TRkAddress.SetField( Index : Integer; Value : string );
begin
  case Index of
    1: FEdtFirstName.DataField := Value;
    2: FEdtLastName.DataField := Value;
    3: FEdtStreet.DataField := Value;
    4: FEdtCity.DataField := Value;
    5: FCbxState.DataField := Value;
    6: FEdtZIP.DataField := Value;
  end;
end;


function TRkAddress.GetCaption( Index : Integer ) : TCaption;
begin
  case Index of
    1: Result := FLblFirstName.Caption;
    2: Result := FLblLastName.Caption;
    3: Result := FLblStreet.Caption;
    4: Result := FLblCity.Caption;
    5: Result := FLblState.Caption;
    6: Result := FLblZIP.Caption;
  end;
end;


procedure TRkAddress.SetCaption( Index : Integer; Value : TCaption );
begin
  case Index of
    1: FLblFirstName.Caption := Value;
    2: FLblLastName.Caption := Value;
    3: FLblStreet.Caption := Value;
    4: FLblCity.Caption := Value;
    5: FLblState.Caption := Value;
    6: FLblZIP.Caption := Value;
  end;
end;


function TRkAddress.GetFontHeight( Font : TFont ) : Integer;
var
  Canvas : TCanvas;
begin
  Canvas := TCanvas.Create;
  try
    Canvas.Handle := GetDC( 0 );
    try
      Canvas.Font := Font;
      Result := Canvas.TextHeight( 'Yy' );
    finally
      ReleaseDC( 0, Canvas.Handle );
    end;
  finally
    Canvas.Free;
  end;
end;


procedure TRkAddress.PositionControls;
var
  FldH : Integer;
  LblH : Integer;
  Line2, Line3 : Integer;
begin
  LblH := GetFontHeight( Font );
  FldH := LblH + GetSystemMetrics( sm_CyBorder ) * 4 + 4;

  if FLabelPosition = lpOnLeft then
  begin
    Line2 := FldH + 4;
    Line3 := 2 * Line2;

    // Arrange fields and labels so that labels are located to the
    // left of each field

    FLblFirstName.SetBounds( 0, 2, 65, LblH  );
    FEdtFirstName.SetBounds( 65, 0, 105, FldH );

    FLblLastName.SetBounds( 170, 2, 63, LblH );
    FLblLastName.Alignment := taRightJustify;
    FEdtLastName.SetBounds( 235, 0, 140, FldH );

    FLblStreet.SetBounds( 0, Line2 + 2, 65, LblH );
    FEdtStreet.SetBounds( 65, Line2, 310, FldH );

    FLblCity.SetBounds( 0, Line3 + 2, 65, LblH );
    FEdtCity.SetBounds( 65, Line3, 125, FldH );

    FLblState.SetBounds( 190, Line3 + 2, 43, LblH );
    FLblState.Alignment := taRightJustify;
    FCbxState.SetBounds( 235, Line3, 60, FldH );

    FLblZIP.SetBounds( 295, Line3 + 2, 28, LblH );
    FLblZIP.Alignment := taRightJustify;
    FEdtZIP.SetBounds( 325, Line3, 50, FldH );
  end
  else
  begin
    Line2 := LblH + FldH + 2;
    Line3 := 2 * Line2;

    // Arrange fields and labels so that labels are located above
    // each field

    FLblFirstName.SetBounds( 0, 0, 120, LblH );
    FEdtFirstName.SetBounds( 0, LblH, 120, FldH );

    FLblLastName.SetBounds( 130, 0, 150, LblH );
    FLblLastName.Alignment := taLeftJustify;
    FEdtLastName.SetBounds( 130, LblH, 150, FldH );

    FLblStreet.SetBounds( 0, Line2, 280, LblH );
    FEdtStreet.SetBounds( 0, Line2 + LblH, 280, FldH );

    FLblCity.SetBounds( 0, Line3, 121, LblH );
    FEdtCity.SetBounds( 0, Line3 + LblH, 121, FldH );

    FLblState.SetBounds( 130, Line3, 70, LblH );
    FLblState.Alignment := taLeftJustify;
    FCbxState.SetBounds( 130, Line3 + LblH, 70, FldH );

    FLblZIP.SetBounds( 210, Line3, 70, LblH );
    FLblZIP.Alignment := taLeftJustify;
    FEdtZIP.SetBounds( 210, Line3 + LblH, 70, FldH );
  end;
  Resize;
end; {= TRkAddress.PositionControls =}


procedure TRkAddress.SetLabelPosition( Value : TLabelPosition );
begin
  if FLabelPosition <> Value then
  begin
    FLabelPosition := Value;
    PositionControls;
  end;
end;


procedure TRkAddress.WMSize( var Msg : TWMSize );
begin
  inherited;
  Resize;
end;


procedure TRkAddress.Resize;
var
  FldH : Integer;
  LblH : Integer;
begin
  LblH := GetFontHeight( Font );
  FldH := LblH + GetSystemMetrics( sm_CyBorder ) * 4 + 4;

  if FLabelPosition = lpOnLeft then
  begin
    Width := 375;
    Height := 3 * ( FldH + 4 );
  end
  else
  begin
    Width := 280;
    Height := 3 * LblH + 3 * FldH + 4;
  end;
end;


procedure TRkAddress.CMFontChanged( var Msg : TMessage );
begin
  inherited;
  // If the font changes, update the position and size of the controls
  PositionControls;
end;


end.
